# Copyright 2026 Oliver Smith, Paul Adam
# SPDX-License-Identifier: GPL-3.0-or-later
from getpass import getpass
from pathlib import Path

import pmb.config
import pmb.install
import pmb.parse
from pmb.core.context import get_context
from pmb.helpers import logging
from pmb.helpers.exceptions import NonBugError


def install(
    add: str,
    android_recovery_zip: bool,
    cipher: str,
    cmdline: str | None,
    filesystem: str,
    full_disk_encryption: bool,
    disk: Path | None,
    install_base: bool,
    install_cgpt: bool,
    install_local_pkgs: bool,
    install_recommends: bool,
    iter_time: str,
    no_fde: bool,
    no_firewall: bool,
    no_image: bool,
    no_reboot: bool,
    no_sshd: bool,
    partition: str | None,
    password: str,
    recovery_flash_kernel: bool,
    recovery_install_partition: str,
    resume: bool,
    rsync: bool,
    sector_size: int | None,
    single_partition: bool,
    sparse: bool | None,
    split: bool | None,
    verbose: bool,
    zap: bool,
) -> None:
    config = get_context().config
    device = config.device
    deviceinfo = pmb.parse.deviceinfo(device)
    is_split = split if split is not None else False
    if no_fde:
        logging.warning("WARNING: --no-fde is deprecated, as it is now the default.")
    if rsync and full_disk_encryption:
        raise ValueError("Installation using rsync is not compatible with full disk encryption.")
    if rsync and not disk:
        raise ValueError("Installation using rsync only works with --disk.")

    if rsync and filesystem == "btrfs":
        raise ValueError("Installation using rsync is not currently supported on btrfs filesystem.")

    if not disk and split is None:
        # Default to split if the flash method requires it
        flasher = pmb.config.flashers.get(deviceinfo.flash_method, {})
        if flasher.get("split", False):
            is_split = True

    # Android recovery zip related
    if android_recovery_zip and filesystem:
        raise ValueError(
            "--android-recovery-zip cannot be combined with --filesystem (patches welcome)"
        )
    if android_recovery_zip and full_disk_encryption:
        logging.info(
            "WARNING: --fde is rarely used in combination with"
            " --android-recovery-zip. If this does not work, consider"
            " using another method (e.g. installing via netcat)"
        )
        logging.info(
            "WARNING: the kernel of the recovery system (e.g. TWRP)"
            f" must support the cryptsetup cipher '{cipher}'."
        )
        logging.info(
            "If you know what you are doing, consider setting a"
            " different cipher with 'pmbootstrap install --cipher=..."
            " --fde --android-recovery-zip'."
        )

    # Don't install locally compiled packages and package signing keys
    if not install_local_pkgs:
        # Implies that we don't build outdated packages (overriding the answer
        # in 'pmbootstrap init')
        config.build_pkgs_on_install = False

        # Safest way to avoid installing local packages is having none
        if list((config.work / "packages").glob("*")):
            raise ValueError(
                "--no-local-pkgs specified, but locally built"
                " packages found. Consider 'pmbootstrap zap -p'"
                " to delete them."
            )

    if not password:
        password = ""
        while not password:
            password = getpass(f"Choose a password for the user '{config.user}': ")
            if not password:
                logging.error("ERROR: Password must not be empty!")

        if getpass(f"Confirm password for '{config.user}': ") != password:
            raise NonBugError("Passwords did not match!")

    # Verify that the root filesystem is supported by current pmaports branch
    pmb.install.get_root_filesystem(filesystem)

    pmb.install.install(
        add,
        android_recovery_zip,
        cipher,
        cmdline,
        filesystem,
        full_disk_encryption,
        disk,
        install_base,
        install_cgpt,
        install_local_pkgs,
        install_recommends,
        iter_time,
        no_firewall,
        no_image,
        no_reboot,
        no_sshd,
        partition,
        password,
        recovery_flash_kernel,
        recovery_install_partition,
        resume,
        rsync,
        sector_size,
        single_partition,
        sparse,
        verbose,
        zap,
        is_split,
    )



# Copyright 2024 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import pmb.helpers.ui
from pmb.core import Config
from pmb.core.config import ServiceManagerConfig
from pmb.helpers.exceptions import NonBugError
from pmb.types import WithExtraRepos


def service_managers_from_packaging(
    ui: str,
) -> tuple[ServiceManagerConfig, list[ServiceManagerConfig], str]:
    """
    Get default and available service managers and reasoning based on the
    selected UI package (and in the future device too).
    """
    default = ServiceManagerConfig.SYSTEMD
    available = [ServiceManagerConfig.SYSTEMD, ServiceManagerConfig.OPENRC]
    reason = ""

    # Select based on UI APKBUILD options:
    # * pmb:support-systemd
    # * pmb:support-openrc
    # * pmb:default-systemd
    # * pmb:default-openrc
    pkgname_ui = f"postmarketos-ui-{ui}"
    apkbuild_ui = pmb.helpers.pmaports.get(pkgname_ui, False, False, WithExtraRepos.DISABLED)
    if not apkbuild_ui:
        default = ServiceManagerConfig.OPENRC
        available = [ServiceManagerConfig.OPENRC]
        reason = f"{pkgname_ui} does not exist on current pmaports branch"
    else:
        opts = apkbuild_ui["options"]
        if "pmb:default-systemd" not in opts and "pmb:default-openrc" not in opts:
            # Legacy case where we don't know whether systemd and openrc are
            # explicitly supported or not. We only know the default by looking
            # at the deprecated pmb:systemd being present or not.
            # TODO: Remove after enforcing in pmaports CI that all UIs have
            # either pmb:default-systemd or pmb:default-openrc set and that all
            # supported pmaports branches have this
            if "pmb:systemd" in opts:
                reason = f"{pkgname_ui} defaults to systemd"
            else:
                default = ServiceManagerConfig.OPENRC
                reason = f"{pkgname_ui} defaults to openrc"
        elif "pmb:support-systemd" not in opts:
            default = ServiceManagerConfig.OPENRC
            available = [ServiceManagerConfig.OPENRC]
            reason = f"{pkgname_ui} doesn't support systemd"
        elif "pmb:support-openrc" not in opts:
            available = [ServiceManagerConfig.SYSTEMD]
            reason = f"{pkgname_ui} doesn't support openrc"
        else:
            # pmb:support-systemd and pmb:support-openrc are in opts
            # as enforced by pmb.helpers.ui.validate_ui_options().
            if "pmb:default-systemd" in opts:
                reason = f"{pkgname_ui} defaults to systemd"
            elif "pmb:default-openrc" in opts:
                default = ServiceManagerConfig.OPENRC
                reason = f"{pkgname_ui} defaults to openrc"
            else:
                raise NonBugError(f"{pkgname_ui} doesn't set a default service manager")

    return default, available, reason


def service_managers_from_user_selection(config: Config) -> tuple[ServiceManagerConfig, str]:
    """
    Get selected service manager and reasoning based on user selection in
    "pmbootstrap init".
    """
    default, available, reason = service_managers_from_packaging(config.ui)
    selected = ServiceManagerConfig.SYSTEMD

    if len(available) > 1:
        match config.service_manager:
            case ServiceManagerConfig.DEFAULT:
                selected = default
                reason = f"default selected in pmbootstrap init, {reason}"
            case ServiceManagerConfig.SYSTEMD:
                selected = ServiceManagerConfig.SYSTEMD
                reason = "systemd selected in pmbootstrap init"
            case ServiceManagerConfig.OPENRC:
                selected = ServiceManagerConfig.OPENRC
                reason = "openrc selected in pmbootstrap init"
    else:
        selected = default

    return selected, reason


def is_systemd_selected(config: Config) -> bool:
    selected, _ = service_managers_from_user_selection(config)
    return selected == ServiceManagerConfig.SYSTEMD


def systemd_selected_str(config: Config) -> tuple[str, str]:
    selected, reasoning = service_managers_from_user_selection(config)
    yesno = "yes" if selected == ServiceManagerConfig.SYSTEMD else "no"
    return yesno, reasoning



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
"""
Functions that work with both pmaports and binary package repos.

See also:

    - pmb/helpers/pmaports.py (work with pmaports)

    - pmb/helpers/repo.py (work with binary package repos)
"""

from typing import overload

import pmb.helpers.pmaports
import pmb.helpers.repo
import pmb.parse.version
from pmb.core.arch import Arch
from pmb.core.package_metadata import PackageMetadata
from pmb.helpers import logging
from pmb.meta import Cache


def remove_operators(package: str) -> str:
    for operator in [">", ">=", "<=", "=", "<", "~"]:
        if operator in package:
            package = package.split(operator)[0]
            break
    return package


def check_version_constraints(pkgname_with_op: str, version: str) -> bool:
    # Operators and matching return values from pmb.parse.version.compare()
    operators = {
        ">=": [0, 1],
        ">": [1],
        "<=": [-1, 0],
        "<": [-1],
    }

    for op, valid_results in operators.items():
        if op not in pkgname_with_op:
            continue
        ver_req = pkgname_with_op.split(op, 1)[1]
        result = pmb.parse.version.compare(version, ver_req)
        logging.verbose(
            f"check_version_constraints: op:{op}, version:{version}, ver_req:{ver_req}, result:{result}"
        )
        if result in valid_results:
            logging.debug(f"{pkgname_with_op}: matches package")
            return True
        else:
            logging.debug(f"{pkgname_with_op}: does not match package")
            return False

    logging.debug(f"check_version_constraints: ignoring {pkgname_with_op}")
    return True


@overload
def get(pkgname: str, arch: Arch) -> PackageMetadata: ...


@overload
def get(pkgname: str, arch: Arch, must_exist: bool = ...) -> PackageMetadata | None: ...


@Cache("pkgname", "arch")
def get(
    pkgname: str,
    arch: Arch,
    must_exist: bool = True,
) -> PackageMetadata | None:
    """
    Find a package in pmaports, and as fallback in the APKINDEXes of the binary packages.

    :param pkgname: package name (e.g. "hello-world")
    :param arch: architecture of the binary package.
    :param must_exist: raise an exception, if not found

    :returns: * data from the parsed APKBUILD or APKINDEX in the following format:
                    {"arch": ["noarch"], "depends": ["busybox-extras", "lddtree", ...],
                    "pkgname": "postmarketos-mkinitfs", "provides": ["mkinitfs=0..1"],
                    "version": "0.0.4-r10"}

        * None if the package was not found
    """
    # Find in pmaports
    pmaport = pmb.helpers.pmaports.get(pkgname, False)
    if pmaport and pmb.helpers.pmaports.check_arches(pmaport["arch"], arch):
        return PackageMetadata.from_pmaport(pmaport)

    # Find in APKINDEX
    pmb.helpers.repo.update(arch)
    ret_repo = pmb.parse.apkindex.package(pkgname, arch, False)
    if ret_repo:
        return PackageMetadata.from_apkindex_block(ret_repo)

    # Could not find the package
    if not must_exist:
        return None
    raise RuntimeError(f"Package '{pkgname}': Could not find it in pmaports or any APKINDEX!")


@Cache("pkgname", "arch")
def depends_recurse(pkgname: str, arch: Arch) -> list[str]:
    """
    Recursively resolve all of the package's dependencies.

    :param pkgname: name of the package (e.g. "device-samsung-i9100")
    :param arch: preferred architecture for binary packages
    :returns: a list of pkgname_start and all its dependencies, e.g:
        ["busybox-static-armhf", "device-samsung-i9100",
        "linux-samsung-i9100", ...]
    """
    # Build ret (by iterating over the queue)
    queue = {pkgname}
    ret: set[str] = set()
    alternative_names: set[str] = set()
    while len(queue):
        pkgname = queue.pop()
        package = get(pkgname, arch)

        # Add its depends to the queue
        for depend in package.depends:
            if depend not in ret and depend not in alternative_names:
                queue.add(depend)

        # Add the pkgname (not possible subpkgname) to ret
        ret.add(package.pkgname)
        alternative_names.update(package.provides)
        alternative_names.add(pkgname)

    return sorted(ret)



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

import copy
import inspect
import os
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Any

import pmb.config
import pmb.helpers.devices
from pmb.core.arch import Arch
from pmb.core.context import get_context
from pmb.core.pkgrepo import pkgrepo_default_path
from pmb.helpers import logging
from pmb.helpers.exceptions import NonBugError
from pmb.helpers.toml import load_toml_file
from pmb.meta import Cache


class InitfsCompressionFormat(Enum):
    ZSTD = "zstd"
    LZ4 = "lz4"
    LZMA = "lzma"
    GZIP = "gzip"
    NONE = "none"

    @staticmethod
    def from_str(compression_format: str) -> InitfsCompressionFormat:
        try:
            return InitfsCompressionFormat(compression_format)
        except ValueError as exception:
            raise ValueError(f"Invalid compression format '{compression_format}'") from exception


class InitfsCompressionLevel(Enum):
    DEFAULT = "default"
    FAST = "fast"
    BEST = "best"

    @staticmethod
    def from_str(compression_level: str) -> InitfsCompressionLevel:
        try:
            return InitfsCompressionLevel(compression_level)
        except ValueError as exception:
            raise ValueError(f"Invalid compression level '{compression_level}'") from exception


@dataclass
class InitfsCompression:
    format_: InitfsCompressionFormat
    level: InitfsCompressionLevel | None

    @staticmethod
    def from_str(initfs_compression: str) -> InitfsCompression:
        segments = initfs_compression.split(":", maxsplit=1)

        # If we can't even figure out the format, the other information is no use. As such, let this
        # raise an exception if it happens.
        format_ = InitfsCompressionFormat.from_str(segments[0])

        try:
            level = InitfsCompressionLevel.from_str(segments[1]) if len(segments) == 2 else None
        except ValueError:
            level = None

        return InitfsCompression(format_, level)


# FIXME: It feels weird to handle this at parse time.
# we should instead have the Deviceinfo object store
# the attributes for all kernels and require the user
# to specify which one they're using.
# Basically: treat Deviceinfo as a standalone type that
# doesn't need to traverse pmaports.
def _parse_kernel_suffix(info: dict[str, str], device: str, kernel: str | None) -> dict[str, str]:
    """
    Remove the kernel suffix (as selected in 'pmbootstrap init') from
    deviceinfo variables. Related:
    https://wiki.postmarketos.org/wiki/Device_specific_package#Multiple_kernels

    :param info: deviceinfo dict, e.g.:
                 {"a": "first",
                  "b_mainline": "second",
                  "b_downstream": "third"}
    :param device: which device info belongs to
    :param kernel: which kernel suffix to remove (e.g. "mainline")
    :returns: info, but with the configured kernel suffix removed, e.g:
              {"a": "first",
               "b": "second",
               "b_downstream": "third"}
    """
    # FIXME: We shouldn't have test-specific code in regular implementations like this.
    # We don't support parsing the kernel variants in tests yet, since this code
    # depends on pmaports being available and calls into a whole lot of other code.
    if os.environ.get("PYTEST_CURRENT_TEST", "").startswith("pmb/parse/test_deviceinfo.py"):
        # If you hit this, you're probably trying to add a test for kernel variants.
        # You'll need to figure out how to mock the APKBUILD parsing below.
        if kernel is not None:
            raise ValueError
        return info
    # Do nothing if the configured kernel isn't available in the kernel (e.g.
    # after switching from device with multiple kernels to device with only one
    # kernel)
    kernels = pmb.parse._apkbuild.kernels(device)
    if not kernels or kernel not in kernels:
        logging.verbose(f"parse_kernel_suffix: {kernel} not in {kernels}")
        return info

    ret = copy.copy(info)

    suffix_kernel = kernel.replace("-", "_")
    for key in inspect.get_annotations(Deviceinfo):
        key_kernel = f"{key}_{suffix_kernel}"
        if key_kernel not in ret:
            continue

        # Move ret[key_kernel] to ret[key]
        logging.verbose(f"parse_kernel_suffix: {key_kernel} => {key}")
        ret[key] = ret[key_kernel]
        del ret[key_kernel]

    return ret


@Cache("device", "kernel")
def deviceinfo(device: str | None = None, kernel: str | None = None) -> Deviceinfo:
    """
    :param device: defaults to args.device
    :param kernel: defaults to args.kernel
    """
    context = get_context()
    if not device:
        device = context.config.device
    if not kernel:
        kernel = context.config.kernel

    path = pmb.helpers.devices.find_path(device, "deviceinfo")
    if not path:
        raise NonBugError(
            f"Device '{device}' not found. Run 'pmbootstrap init' to start a new device port or to "
            "choose another device. It may have been renamed, see <https://postmarketos.org/renamed>"
        )

    return Deviceinfo(path, kernel)


class Deviceinfo:
    """Variables from deviceinfo. Reference: <https://postmarketos.org/deviceinfo>"""

    path: Path
    # general
    format_version: str
    name: str
    manufacturer: str
    codename: str
    year: str
    dtb: str = ""
    arch: Arch

    # device
    chassis: str | None
    keyboard: str | None = ""  # deprecated
    drm: bool | None = False
    dev_touchscreen: str | None = ""
    dev_touchscreen_calibration: str | None = ""
    append_dtb: str | None = ""

    # bootloader
    flash_method: str = ""
    boot_filesystem: str | None = ""
    create_initfs_extra: bool | None = False
    create_prep_boot: bool | None = False
    initfs_compression: InitfsCompression = InitfsCompression(InitfsCompressionFormat.GZIP, None)

    # flash
    flash_heimdall_partition_kernel: str | None = ""
    flash_heimdall_partition_vendor_boot: str | None = ""
    flash_heimdall_partition_initfs: str | None = ""
    flash_heimdall_partition_rootfs: str | None = ""
    flash_heimdall_partition_system: str | None = ""  # deprecated
    flash_heimdall_partition_vbmeta: str | None = ""
    flash_heimdall_partition_dtbo: str | None = ""
    flash_fastboot_partition_kernel: str | None = ""
    flash_fastboot_partition_vendor_boot: str | None = ""
    flash_fastboot_partition_rootfs: str | None = ""
    flash_fastboot_partition_system: str | None = ""  # deprecated
    flash_fastboot_partition_vbmeta: str | None = ""
    flash_fastboot_partition_dtbo: str | None = ""
    flash_rk_partition_kernel: str | None = ""
    flash_rk_partition_vendor_boot: str | None = ""
    flash_rk_partition_rootfs: str | None = ""
    flash_rk_partition_system: str | None = ""  # deprecated
    flash_mtkclient_partition_kernel: str | None = ""
    flash_mtkclient_partition_vendor_boot: str | None = ""
    flash_mtkclient_partition_rootfs: str | None = ""
    flash_mtkclient_partition_vbmeta: str | None = ""
    flash_mtkclient_partition_dtbo: str | None = ""
    generate_legacy_uboot_initfs: str | None = ""
    kernel_cmdline: str | None = ""
    generate_bootimg: str | None = ""
    header_version: int | None = None
    bootimg_qcdt: str | None = ""
    bootimg_qcdt_type: str | None = ""
    bootimg_mtk_mkimage: str | None = ""  # deprecated
    bootimg_mtk_label_kernel: str | None = ""
    bootimg_mtk_label_ramdisk: str | None = ""
    bootimg_dtb_second: str | None = ""
    bootimg_custom_args: str | None = ""
    flash_offset_base: str | None = ""
    flash_offset_dtb: str | None = ""
    flash_offset_kernel: str | None = ""
    flash_offset_ramdisk: str | None = ""
    flash_offset_second: str | None = ""
    flash_offset_tags: str | None = ""
    flash_pagesize: str | None = ""
    flash_fastboot_max_size: str | None = ""
    flash_sparse: str | None = ""
    flash_sparse_samsung_format: str | None = ""
    rootfs_image_sector_size: str | None = ""
    sd_embed_firmware: str | None = ""
    sd_embed_firmware_step_size: str | None = ""
    partition_blacklist: str | None = ""
    boot_part_start: str | None = ""
    partition_type: str | None = ""
    root_filesystem: str | None = ""
    flash_kernel_on_update: str | None = ""
    cgpt_kpart: str | None = ""
    cgpt_kpart_start: str | None = ""
    cgpt_kpart_size: str | None = ""

    # weston
    weston_pixman_type: str | None = ""

    # keymaps
    keymaps: str | None = ""

    def __init__(self, path: Path, kernel: str | None = None) -> None:
        ret = {}
        with open(path) as handle:
            for line in handle:
                if not line.startswith("deviceinfo_"):
                    continue
                if "=" not in line:
                    raise SyntaxError(f"{path}: No '=' found:\n\t{line}")
                split = line.split("=", 1)
                key = split[0][len("deviceinfo_") :]
                value = split[1].replace('"', "").replace("\n", "")
                ret[key] = value

        ret = _parse_kernel_suffix(ret, ret["codename"], kernel)

        for key, value in ret.items():
            # FIXME: something to turn on and fix in the future
            # Should be rewritten to use deviceinfo_schema.toml as source of truth instead of the
            # Deviceinfo class keys first though.
            # if key not in Deviceinfo.__annotations__.keys():
            #     logging.warning(f"deviceinfo: {key} is not a known attribute")
            match key:
                case "arch":
                    self.arch = Arch.from_str(value)
                case "gpu_accelerated":  # deprecated
                    self.drm = value == "true"
                case "header_version":
                    self.header_version = int(value)
                case "initfs_compression":
                    self.initfs_compression = InitfsCompression.from_str(value)
                case _:
                    setattr(self, key, value)

        if not self.flash_method:
            self.flash_method = "none"


class DeviceinfoSchemaDatatype(Enum):
    INTEGER = "integer"
    ENUMERATION = "enumeration"
    BOOLEAN = "boolean"
    STRING = "string"

    @property
    def default_value_type(self) -> type:
        match self:
            case DeviceinfoSchemaDatatype.INTEGER:
                return int
            case DeviceinfoSchemaDatatype.ENUMERATION:
                return str
            case DeviceinfoSchemaDatatype.BOOLEAN:
                return bool
            case DeviceinfoSchemaDatatype.STRING:
                return str


@dataclass
class DeviceinfoSchemaVariable:
    """A variable in the deviceinfo schema."""

    name: str
    description: str
    datatype: DeviceinfoSchemaDatatype = DeviceinfoSchemaDatatype.STRING
    mandatory: bool = False
    allow_device_variant_suffix: bool = False
    behaviour_if_unset: str | None = None
    fate: str | None = None
    # Any following attributes could be expressed more concisely if not for
    # Python's quite limited type system / enum APIs
    epitaph: str | None = None
    integer_interval: range | None = None
    enum_values: set[str] | None = None
    # mypy treats Any | None different from Any
    default_value: Any | None = None

    def validate(self) -> None:
        # This is where we test what Python cannot test for us, i.e. that all
        # attributes are only set if specific conditions are met
        if self.integer_interval and self.datatype != DeviceinfoSchemaDatatype.INTEGER:
            raise NonBugError(
                f"Deviceinfo schema variable deviceinfo_{self.name} does not set integer type but has an integer interval"
            )
        if self.enum_values and self.datatype != DeviceinfoSchemaDatatype.ENUMERATION:
            raise NonBugError(
                f"Deviceinfo schema variable deviceinfo_{self.name} does not set enumeration type but has enum values"
            )
        if self.epitaph and not self.fate:
            raise NonBugError(
                f"Deviceinfo schema variable deviceinfo_{self.name} has an epitaph but does not have a fate set"
            )

        if self.default_value is not None:
            default_value_type = type(self.default_value)
            expected_value_type = self.datatype.default_value_type
            if default_value_type != expected_value_type:
                raise NonBugError(
                    f"Deviceinfo schema variable deviceinfo_{self.name} has default value of type {default_value_type}, expected {expected_value_type}"
                )


@dataclass
class DeviceinfoSchemaCategory:
    """A category in the deviceinfo schema."""

    name: str
    description: str
    variables: dict[str, DeviceinfoSchemaVariable]


@dataclass
class DeviceinfoSchema:
    """Parsed deviceinfo schema."""

    schema_version: str
    categories: dict[str, DeviceinfoSchemaCategory]
    # We're omitting renames for now

    def get(self, category_name: str, variable_name: str) -> DeviceinfoSchemaVariable | None:
        """Look up a variable in a category of the schema"""
        if not (category := self.categories.get(category_name)):
            return None
        return category.variables.get(variable_name)


@Cache()
def deviceinfo_schema() -> DeviceinfoSchema:
    """Parse the deviceinfo_schema.toml file in pmaports into a structured format"""
    deviceinfo_schema_path = pkgrepo_default_path() / "deviceinfo_schema.toml"
    if not deviceinfo_schema_path.exists():
        raise NonBugError(
            "pmaports checkout does not contain a 'deviceinfo_schema.toml'. You may be on an unsupported branch"
        )

    toml = load_toml_file(deviceinfo_schema_path)

    if not (metadata := toml.get("metadata")):
        raise NonBugError("Deviceinfo schema does not contain a 'metadata' section")

    if (schema_version := metadata.get("schema_version")) != "0.1":
        raise NonBugError(
            f"pmbootstrap only implements deviceinfo schema version 0.1, got {schema_version}"
        )

    if not (categories := metadata.get("categories")):
        raise NonBugError("No categories found under the 'metadata' table in the deviceinfo schema")

    if not (variables := toml.get("variable")):
        raise NonBugError("No variables found in the deviceinfo schema")

    # For the sake of simplicity we assume that the file is mostly
    # well-formatted, otherwise this would be extremely verbose

    parsed_categories: dict[str, DeviceinfoSchemaCategory] = {}

    for category_name, category_description in categories.items():
        parsed_variables: dict[str, DeviceinfoSchemaVariable] = {}

        for variable_name, variable in variables[category_name].items():
            # Some of these need type conversions
            if datatype := variable.get("datatype"):
                variable["datatype"] = DeviceinfoSchemaDatatype(datatype)
            if enum_values := variable.get("enum_values"):
                variable["enum_values"] = set(enum_values)
            if interval := variable.get("integer_interval"):
                parts = interval[1:-1].split(", ")
                start = int(parts[0])
                end = int(parts[1])
                # Python ranges are start inclusive by default
                if interval[0] == "(":
                    start += 1
                # ...and end exclusive by default
                if interval[-1] == "]":
                    end += 1
                variable["integer_interval"] = range(start, end)

            parsed_variable = DeviceinfoSchemaVariable(
                name=variable_name,
                **variable,
            )
            parsed_variable.validate()
            parsed_variables[variable_name] = parsed_variable

        parsed_category = DeviceinfoSchemaCategory(
            name=category_name, description=category_description, variables=parsed_variables
        )
        parsed_categories[category_name] = parsed_category

    return DeviceinfoSchema(schema_version=schema_version, categories=parsed_categories)


@Cache()
def deviceinfo_schema_default_boot_filesystem() -> str:
    """Reads the deviceinfo schema and returns the default boot filesystem specified"""
    variable = deviceinfo_schema().get("flash", "boot_filesystem")
    if variable and variable.default_value is not None:
        return variable.default_value
    elif variable:
        raise NonBugError(
            "[variable.flash.boot_filesystem] lacks a default value in the deviceinfo schema"
        )
    else:
        raise NonBugError(
            "Could not find [variable.flash.boot_filesystem] in the deviceinfo schema"
        )



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import pmb.chroot.other
from pmb.core import Chroot, ChrootType


def _parse_flavor(device: str, autoinstall: bool = True) -> str:
    """
    Verify the flavor argument if specified, or return a default value.

    :param autoinstall: make sure that at least one kernel flavor is installed
    """
    # Install a kernel and get its "flavor", where flavor is a pmOS-specific
    # identifier that is typically in the form
    # "postmarketos-<manufacturer>-<device/chip>", e.g.
    # "postmarketos-qcom-sdm845"
    chroot = Chroot(ChrootType.ROOTFS, device)
    flavor = pmb.chroot.other.kernel_flavor_installed(chroot, autoinstall)

    if not flavor:
        raise RuntimeError(
            f"No kernel flavors installed in chroot '{chroot}'! Please let"
            " your device package depend on a package starting with 'linux-'."
        )
    return flavor
